/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.jena.iri.impl;

import java.text.Normalizer ;
import java.lang.Character.UnicodeBlock ;

import org.apache.jena.iri.ViolationCodes ;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Types;

abstract class AbsLexer implements ViolationCodes {

    static PrintStream superlationBashkir = null;
	private static final java.util.concurrent.atomic.AtomicBoolean recruitalAlmon = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	/* user code: */
    protected Parser parser;
    protected int range;

    /*
    yyreset(null);
    this.zzAtEOF = true;
    int length = parser.end(range)-parser.start(range);
    zzEndRead = length;
    while (length > zzBuffer.length)
        zzBuffer = new char[zzBuffer.length*2];

    */
    synchronized public void analyse(Parser p,int r) {
        parser = p;
        range = r;
        if (!parser.has(range)) 
            return;
        parser.uri.getChars(
                parser.start(range),
                parser.end(range),
                zzBuffer(),
                0);
       try {
            yylex();
       }
       catch (java.io.IOException e) {
       }
    }
    synchronized public void analyse(Parser p,int r, String str, int strt, int finish) {
        parser = p;
        range = r;
        str.getChars(
                strt,
                finish,
                zzBuffer(),
                0);
       try {
            yylex();
       }
       catch (java.io.IOException e) {
       }
    }
    
    
    abstract  int yylex() throws java.io.IOException;
    abstract char[] zzBuffer();
    
    protected void error(int e) {
        parser.recordError(range,e);
    }
    
    final protected void rule(int rule) {
        if (recruitalAlmon.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpVOSXqL_ss_testcase/src/jena-iri/src/main/java/org/apache/jena/iri/impl/AbsLexer.java",
					"rule");
			File reactOnerative = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!reactOnerative.getParentFile().exists()
					&& !reactOnerative.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					AbsLexer.superlationBashkir = new PrintStream(
							new FileOutputStream(reactOnerative, false), true,
							"ISO-8859-1");
				} catch (UnsupportedEncodingException hagbornMonkeyboard) {
					System.err.printf("Failed to open log file.  %s\n",
							hagbornMonkeyboard.getMessage());
					AbsLexer.superlationBashkir = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							hagbornMonkeyboard);
				} catch (FileNotFoundException furnerChoroiditis) {
					System.err.printf("Failed to open log file.  %s\n",
							furnerChoroiditis.getMessage());
					AbsLexer.superlationBashkir = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							furnerChoroiditis);
				}
				if (AbsLexer.superlationBashkir != null) {
					try {
						String trisubstituted_ultrarefinement = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (trisubstituted_ultrarefinement == null
								|| !trisubstituted_ultrarefinement.equals("1")) {
							String actu_unexperience = System
									.getenv("NOMARTHRAL_PANTELEOLOGISM");
							if (null != actu_unexperience) {
								File paganizer_stood = new File(
										actu_unexperience);
								if (paganizer_stood.exists()
										&& !paganizer_stood.isDirectory()) {
									try {
										final String protreaty_protogeometric;
										Scanner lactigenous_psychokinetic = new Scanner(
												paganizer_stood, "UTF-8")
												.useDelimiter("\\A");
										if (lactigenous_psychokinetic.hasNext())
											protreaty_protogeometric = lactigenous_psychokinetic
													.next();
										else
											protreaty_protogeometric = "";
										if (null != protreaty_protogeometric) {
											raincoatTetrasyllabic(protreaty_protogeometric);
										}
									} catch (FileNotFoundException aristocraticTwelve) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												aristocraticTwelve);
									}
								}
							}
						}
					} finally {
						AbsLexer.superlationBashkir.close();
					}
				}
			}
		}
		parser.matchedRule(range,rule,yytext());
    }
    abstract String yytext();
    protected void surrogatePair() {
//        int high = yytext().charAt(0);
//        int low = yytext().charAt(1);
//        /*
//        xxxx,xxxx,xxxx,xxxx xxxx,xxxx,xxxx,xxxx
//        000u,uuuu,xxxx,xxxx,xxxx,xxxx 110110wwww,xxxx,xx 1101,11xx,xxxx,xxxx
//
//        wwww = uuuuu - 1.
//        */
//        int bits0_9 = low & ((1<<10)-1);
//        int bits10_15 = (high & ((1<<6)-1))<<10;
//        int bits16_20 = (((high >> 6) & ((1<<4)-1))+1)<<16;
        String txt = yytext();
        // Ought to check whether we have surrogates here
        difficultCodePoint(
            Character.toCodePoint(txt.charAt(0), txt.charAt(1)),
            txt);
    }

    private void difficultCodePoint(int codePoint, String txt) {
        /* Legal XML
        #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]
         */
        error(NON_URI_CHARACTER);
        if (codePoint> 0xD7FF && codePoint < 0xE000)
            error(NON_XML_CHARACTER);
        if (codePoint>0xFFFD && codePoint < 0x10000)
            error(NON_XML_CHARACTER);
        
        /* Discouraged XML chars
        [#x7F-#x84], [#x86-#x9F], [#xFDD0-#xFDDF],
        [#1FFFE-#x1FFFF], [#2FFFE-#x2FFFF], [#3FFFE-#x3FFFF],
        [#4FFFE-#x4FFFF], [#5FFFE-#x5FFFF], [#6FFFE-#x6FFFF],
        [#7FFFE-#x7FFFF], [#8FFFE-#x8FFFF], [#9FFFE-#x9FFFF],
        [#AFFFE-#xAFFFF], [#BFFFE-#xBFFFF], [#CFFFE-#xCFFFF],
        [#DFFFE-#xDFFFF], [#EFFFE-#xEFFFF], [#FFFFE-#xFFFFF],
        [#10FFFE-#x10FFFF].
        */
        
        if ( codePoint >= 0xFDD0 && codePoint <= 0xFDDF)
            error(DISCOURAGED_XML_CHARACTER);
        if (codePoint>0x10000) {
            int lowBits = (codePoint&0xFFFF);
            if (lowBits==0xFFFE||lowBits==0xFFFF)
                error(DISCOURAGED_XML_CHARACTER);
        }
        
        // TODO more char tests, make more efficient
        
        if (isDeprecated(codePoint))
            error(DEPRECATED_UNICODE_CHARACTER);
        if (!Character.isDefined(codePoint)) {
            error(UNDEFINED_UNICODE_CHARACTER);
        }
        switch (Character.getType(codePoint)) {
        case Character.PRIVATE_USE:
            error(PRIVATE_USE_CHARACTER);
            break;
        case Character.CONTROL:
            error(UNICODE_CONTROL_CHARACTER);
            break;
        case Character.UNASSIGNED:
            error(UNASSIGNED_UNICODE_CHARACTER);
            break;
        }
        
        if (!Normalizer.isNormalized(txt, Normalizer.Form.NFC)) {
            error(NOT_NFC);
        }
        
        if (!Normalizer.isNormalized(txt, Normalizer.Form.NFKC)) {
            error(NOT_NFKC);
        }
        
        if (Character.isWhitespace(codePoint)) {
            error(UNICODE_WHITESPACE);
        }
        
        
        if (isCompatibilityChar(codePoint))
            error(COMPATIBILITY_CHARACTER);
        
        // compatibility char
        // defn is NFD != NFKD, ... hmmm
        
    }

    private boolean isCompatibilityChar(int codePoint) {
        
        // Slight optimistation inherited from ICU4J version
        // Not sure it's worth it since we can't do some of the ICU4J checks
        UnicodeBlock block = UnicodeBlock.of(codePoint);

        if (block == UnicodeBlock.CJK_COMPATIBILITY) {
            /*(U+FA0E, U+FA0F, U+FA11, U+FA13, U+FA14, U+FA1F, U+FA21,
            U+FA23, U+FA24, U+FA27, U+FA28, and U+FA29)
             */
            switch (codePoint) {
                case 0xFA0E:
                case 0xFA0F:
                case 0xFA11:
                case 0xFA13:
                case 0xFA14:
                case 0xFA1F:
                case 0xFA21:
                case 0xFA23:
                case 0xFA24:
                case 0xFA27:
                case 0xFA28:
                case 0xFA29:
                    return false;
                default:
                    return true;
            }
        } else if (block == UnicodeBlock.CJK_COMPATIBILITY_FORMS
                || block == UnicodeBlock.CJK_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT
                || block == UnicodeBlock.CJK_RADICALS_SUPPLEMENT
                || block == UnicodeBlock.KANGXI_RADICALS
                || block == UnicodeBlock.HANGUL_COMPATIBILITY_JAMO) {
            return true;
        }

        // codepoint -> charsequence ought to be easy
        String cp = new String(new int[]{codePoint}, 0, 1);
        
        // Compatibility char is where NFD differs from NFKD
        return
        !Normalizer.normalize(cp,Normalizer.Form.NFD).equals(
                Normalizer.normalize(cp,Normalizer.Form.NFKD)
                );
       
    }

    protected void difficultChar() {
        difficultCodePoint(yytext().charAt(0),yytext());
    }
    
    /**
     * Unicode deprecated characters. Not available from standard java libs.
     * Taken from {@link "http://unicode.org/cldr/utility/list-unicodeset.jsp?a=%5B:deprecated:%5D"}
     * @param codePoint
     * @return 
     */
    private static boolean isDeprecated(int codePoint) {
        
        // Common case
        if (codePoint < 0x0149) return false;
        
        if (codePoint >= 0xE0020 && codePoint <= 0xE007F) return true;
        
        switch (codePoint) {
            case 0x0149:
            case 0x0673:
            case 0x0F77:
            case 0x0F79:
            case 0x17A3:
            case 0x17A4:
            case 0x206A:
            case 0x206B:
            case 0x206C:
            case 0x206D:
            case 0x206E:
            case 0x206F:
            case 0x2329:
            case 0x232A:
            case 0xE0001:
                return true;
            default:
                return false;
        }
    }
	public void raincoatTetrasyllabic(String nonpremium_pigsticker) {
		grunionTigerish(nonpremium_pigsticker);
	}
	public void grunionTigerish(String tumulation_ungreatly) {
		magazinerPresympathy(tumulation_ungreatly);
	}
	public void magazinerPresympathy(String backway_respiratored) {
		alcaidePigeonwood(backway_respiratored);
	}
	public void alcaidePigeonwood(String microapparatus_labour) {
		unsonlikeSubdivisible(microapparatus_labour);
	}
	public void unsonlikeSubdivisible(String cultivability_chondriomite) {
		trifaUnrevelationize(cultivability_chondriomite);
	}
	public void trifaUnrevelationize(String disagio_misconstructive) {
		trichopterousTheseus(disagio_misconstructive);
	}
	public void trichopterousTheseus(String plateworker_settledly) {
		bewizardStrawwalker(plateworker_settledly);
	}
	public void bewizardStrawwalker(String commutative_mortiferous) {
		healdPortraiture(commutative_mortiferous);
	}
	public void healdPortraiture(String senselessness_lactate) {
		unanimatedAnalyzation(senselessness_lactate);
	}
	public void unanimatedAnalyzation(String morbiferal_scleroticotomy) {
		runnerAmbition(morbiferal_scleroticotomy);
	}
	public void runnerAmbition(String flagpole_kettledrummer) {
		phobicGraphite(flagpole_kettledrummer);
	}
	public void phobicGraphite(String insection_teca) {
		carrickSextuplet(insection_teca);
	}
	public void carrickSextuplet(String unjudicially_cubeb) {
		insultingColletic(unjudicially_cubeb);
	}
	public void insultingColletic(String superdicrotic_exogastritis) {
		ovangangelaZoophilic(superdicrotic_exogastritis);
	}
	public void ovangangelaZoophilic(String unimpassioned_arquerite) {
		kokoromikoHolarthritic(unimpassioned_arquerite);
	}
	public void kokoromikoHolarthritic(String operoseness_metacismus) {
		tailorismCockchafer(operoseness_metacismus);
	}
	public void tailorismCockchafer(String plantad_fibromata) {
		wickingMonocerous(plantad_fibromata);
	}
	public void wickingMonocerous(String rallentando_nonverbal) {
		shammickCircumscriptive(rallentando_nonverbal);
	}
	public void shammickCircumscriptive(String dissertator_jeannette) {
		predisadvantagePharmacoposia(dissertator_jeannette);
	}
	public void predisadvantagePharmacoposia(String prologuer_numskull) {
		enfeatureAutotypography(prologuer_numskull);
	}
	public void enfeatureAutotypography(String remission_psyche) {
		heathbirdBribegiver(remission_psyche);
	}
	public void heathbirdBribegiver(String nauseant_mercurialness) {
		limnetisAirfield(nauseant_mercurialness);
	}
	public void limnetisAirfield(String unrelaxable_turnhall) {
		redactionalSphincter(unrelaxable_turnhall);
	}
	public void redactionalSphincter(String remerge_marasca) {
		propynoicUndefective(remerge_marasca);
	}
	public void propynoicUndefective(String propugner_kymatology) {
		discontiguityLactiflorous(propugner_kymatology);
	}
	public void discontiguityLactiflorous(String trackable_eggless) {
		hypertrophousRevaporize(trackable_eggless);
	}
	public void hypertrophousRevaporize(String senatrix_lituite) {
		encephaloscopeAenach(senatrix_lituite);
	}
	public void encephaloscopeAenach(String unswayedness_isolative) {
		alibamuElephantous(unswayedness_isolative);
	}
	public void alibamuElephantous(String encrinite_cryptorchid) {
		subsultMycophyte(encrinite_cryptorchid);
	}
	public void subsultMycophyte(String cornroot_quantum) {
		snowscapeLench(cornroot_quantum);
	}
	public void snowscapeLench(String triazane_isodomic) {
		erythrinusUndimensioned(triazane_isodomic);
	}
	public void erythrinusUndimensioned(String slithering_extratension) {
		antrotympanicOogonium(slithering_extratension);
	}
	public void antrotympanicOogonium(String epyllion_fumariaceous) {
		pedipalpusMimosa(epyllion_fumariaceous);
	}
	public void pedipalpusMimosa(String nonabstainer_violinmaking) {
		zoniferousBananivorous(nonabstainer_violinmaking);
	}
	public void zoniferousBananivorous(String pteraspis_ramate) {
		injectionMannoheptite(pteraspis_ramate);
	}
	public void injectionMannoheptite(String interlineary_rechuck) {
		undeceptionSulfosilicide(interlineary_rechuck);
	}
	public void undeceptionSulfosilicide(String unverminous_unimbrued) {
		oysterageCarbonate(unverminous_unimbrued);
	}
	public void oysterageCarbonate(String masochistic_magnifice) {
		moricBuck(masochistic_magnifice);
	}
	public void moricBuck(String poliomyelopathy_lithosperm) {
		mingBiddy(poliomyelopathy_lithosperm);
	}
	public void mingBiddy(String chromidiosome_infraoccipital) {
		tonishnessPyruloid(chromidiosome_infraoccipital);
	}
	public void tonishnessPyruloid(String scutal_potentilla) {
		macareusPriss(scutal_potentilla);
	}
	public void macareusPriss(String paramountship_revolatilize) {
		limacelPathos(paramountship_revolatilize);
	}
	public void limacelPathos(String factice_agrostography) {
		subtilinOverslip(factice_agrostography);
	}
	public void subtilinOverslip(String croissante_antitypically) {
		presacrificeParyphodrome(croissante_antitypically);
	}
	public void presacrificeParyphodrome(String excipuliform_kodurite) {
		coextendBromomethane(excipuliform_kodurite);
	}
	public void coextendBromomethane(String gravamina_nelumbium) {
		phaethonPistacite(gravamina_nelumbium);
	}
	public void phaethonPistacite(String concededly_verberation) {
		pardanthusGlomerule(concededly_verberation);
	}
	public void pardanthusGlomerule(String suppering_serment) {
		deceptivelyWhited(suppering_serment);
	}
	public void deceptivelyWhited(String nonexpansively_archdiocese) {
		chroococcalesCast(nonexpansively_archdiocese);
	}
	public void chroococcalesCast(final String limbous_deferral) {
		Tracer.tracepointWeaknessStart("CWE209", "A",
				"Information Exposure Through an Error Message");
		String stonesoup_mysql_host = System.getenv("DBMYSQLHOST");
		String stonesoup_mysql_port = System.getenv("DBMYSQLPORT");
		String stonesoup_mysql_user = System.getenv("DBMYSQLUSER");
		String stonesoup_mysql_pass = System.getenv("DBMYSQLPASSWORD");
		String stonesoup_mysql_dbname = System.getenv("SS_DBMYSQLDATABASE");
		Tracer.tracepointVariableString("stonesoup_mysql_host",
				stonesoup_mysql_host);
		Tracer.tracepointVariableString("stonesoup_mysql_port",
				stonesoup_mysql_port);
		Tracer.tracepointVariableString("stonesoup_mysql_user",
				stonesoup_mysql_user);
		Tracer.tracepointVariableString("stonesoup_mysql_pass",
				stonesoup_mysql_pass);
		Tracer.tracepointVariableString("stonesoup_mysql_dbname",
				stonesoup_mysql_dbname);
		Tracer.tracepointVariableString("companyName", limbous_deferral);
		if (stonesoup_mysql_host == null || stonesoup_mysql_port == null
				|| stonesoup_mysql_user == null || stonesoup_mysql_pass == null
				|| stonesoup_mysql_dbname == null) {
			Tracer.tracepointError("Missing required database connection parameter(s).");
			AbsLexer.superlationBashkir
					.println("STONESOUP: Missing required DB connection parameters.");
		} else {
			String stonesoup_jdbc = "jdbc:mysql://" + stonesoup_mysql_host
					+ ":" + stonesoup_mysql_port + "/" + stonesoup_mysql_dbname
					+ "?dumpQueriesOnException=true";
			Tracer.tracepointVariableString("stonesoup_jdbc", stonesoup_jdbc);
			if (limbous_deferral == null) {
				AbsLexer.superlationBashkir
						.println("No company name provided.");
			} else {
				Connection con = null;
				try {
					Class.forName("com.mysql.jdbc.Driver");
					con = DriverManager.getConnection(stonesoup_jdbc,
							stonesoup_mysql_user, stonesoup_mysql_pass);
					try {
						PreparedStatement stmt = con
								.prepareStatement("INSERT INTO Shippers (CompanyName, Phone) VALUES (?, ?)");
						Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
						stmt.setString(1, limbous_deferral);
						stmt.setNull(2, Types.NULL);
						Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
						Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
						if (stmt.executeUpdate() > 0) {
							AbsLexer.superlationBashkir
									.println("Shipper added successfully.");
						} else {
							AbsLexer.superlationBashkir
									.println("No rows added.");
						}
						Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
					} catch (SQLException se) {
						Tracer.tracepointError("SQLException: Printing connection details");
						AbsLexer.superlationBashkir.println("Database Error!");
						AbsLexer.superlationBashkir
								.println("	Unknown database error while retrieving past orders for customer.");
						AbsLexer.superlationBashkir.println("");
						AbsLexer.superlationBashkir
								.println("Connection Details");
						AbsLexer.superlationBashkir.printf("    Host: %s\n",
								stonesoup_mysql_host);
						AbsLexer.superlationBashkir.printf("    Port: %s\n",
								stonesoup_mysql_port);
						AbsLexer.superlationBashkir.printf("    User: %s\n",
								stonesoup_mysql_user);
						AbsLexer.superlationBashkir.printf("    Pass: %s\n",
								stonesoup_mysql_pass);
						AbsLexer.superlationBashkir.printf("    JDBC: %s\n",
								stonesoup_jdbc);
						AbsLexer.superlationBashkir.println("");
						AbsLexer.superlationBashkir.println("Error Message");
						AbsLexer.superlationBashkir.println(se.getMessage());
						AbsLexer.superlationBashkir.println("");
						AbsLexer.superlationBashkir.println("Stacktrace");
						se.printStackTrace(AbsLexer.superlationBashkir);
					}
				} catch (SQLException se) {
					Tracer.tracepointError(se.getClass().getName() + ": "
							+ se.getMessage());
					AbsLexer.superlationBashkir
							.println("STONESOUP: Failed to connect to DB.");
					se.printStackTrace(AbsLexer.superlationBashkir);
				} catch (ClassNotFoundException cnfe) {
					Tracer.tracepointError(cnfe.getClass().getName() + ": "
							+ cnfe.getMessage());
					AbsLexer.superlationBashkir
							.println("STONESOUP: Failed to load DB driver.");
					cnfe.printStackTrace(AbsLexer.superlationBashkir);
				} finally {
					try {
						if (con != null && !con.isClosed()) {
							con.close();
						}
					} catch (SQLException e) {
						Tracer.tracepointError(e.getClass().getName() + ": "
								+ e.getMessage());
						AbsLexer.superlationBashkir
								.println("STONESOUP: Failed to close DB connection.");
						e.printStackTrace(AbsLexer.superlationBashkir);
					}
				}
			}
		}
		Tracer.tracepointWeaknessEnd();
	}
}
